/*
 * CifCrystal.cpp
 *
 * Copyright Notice: see copyright.txt
 *
 *  Date: 3/8/2012 (D/M/Y)
 *  Author: Dmitrij Lioubartsev
 *  Email: dmitrijl42@gmail.com
 *
 *  For class info, see CifCrystal.h
 */

#include "CifCrystal.h"
#include "Atom.h"
#include "SymmetryCalculator.h"


#include <iostream>
#include <cmath>
using namespace std;

#define _PI__ 3.14159265

CifCrystal::CifCrystal() {
	//constants
	a = b = c = 0;
	alpha = beta = gamma = 0;
	aNew = bNew = cNew = 0;
	aDup = bDup = cDup = 0;
	for(int i = 0; i < 6; i++) {
		isSet[i] = false;
	}
	//symmetry
	vector<string> symmetries(0);
	//atoms
	vector<Atom> atoms(0);
	numEl = numx = numy = numz = 0;
	vector<Bond> bonds(0);

}

CifCrystal::~CifCrystal() {
	// do nothing
}

bool CifCrystal::seta(double value) {
	if(!isSet[0]) {
		a = value;
		isSet[0] = true;
		return true;
	} else {
		return false;
	}
}

bool CifCrystal::setb(double value) {
	if(!isSet[1]) {
		b = value;
		isSet[1] = true;
		return true;
	} else {
		return false;
	}
}

bool CifCrystal::setc(double value) {
	if(!isSet[2]) {
		c = value;
		isSet[2] = true;
		return true;
	} else {
		return false;
	}
}

bool CifCrystal::setalpha(double value) {
	if(!isSet[3]) {
		alpha = value;
		isSet[3] = true;
		return true;
	} else {
		return false;
	}
}

bool CifCrystal::setbeta(double value) {
	if(!isSet[4]) {
		beta = value;
		isSet[4] = true;
		return true;
	} else {
		return false;
	}
}

bool CifCrystal::setgamma(double value) {
	if(!isSet[5]) {
		gamma = value;
		isSet[5] = true;
		return true;
	} else {
		return false;
	}
}

//returns an array of size 6 to indicate which values has been set.
// order is {a, b, c, alpha, beta, gamma}
bool* CifCrystal::areSet() { return isSet;}

double CifCrystal::get_a() { return a;}
double CifCrystal::get_b() { return b;}
double CifCrystal::get_c() { return c;}
double CifCrystal::get_alpha() { return alpha;}
double CifCrystal::get_beta() {	return beta;}
double CifCrystal::get_gamma() { return gamma;}
double CifCrystal::getNewa() { return aNew;}
double CifCrystal::getNewb() { return bNew;}
double CifCrystal::getNewc() { return cNew;}
int CifCrystal::getDupa() { return aDup;}
int CifCrystal::getDupb() { return bDup;}
int CifCrystal::getDupc() { return cDup;}

/*
 * Adds a symmetry
 */
void CifCrystal::addSymmetry(string symmetry) {
	symmetries.push_back(symmetry);
}

//atoms

//setters, see CifCrystal.h for more info
void CifCrystal::setAx(double x) {
	if(numx < atoms.size()) {
		atoms[numx].setx(x);
	} else {
		Atom a = Atom();
		a.setx(x);
		atoms.push_back(a);
	}
	numx++;
}

void CifCrystal::setAy(double y) {
	if(numy < atoms.size()) {
		atoms[numy].sety(y);
	} else {
		Atom a = Atom();
		a.sety(y);
		atoms.push_back(a);
	}
	numy++;
}

void CifCrystal::setAz(double z) {
	if(numz < atoms.size()) {
		atoms[numz].setz(z);
	} else {
		Atom a = Atom();
		a.setz(z);
		atoms.push_back(a);
	}
	numz++;
}

void CifCrystal::setAElement(string e) {
	if(numEl < atoms.size()) {
		atoms[numEl].setElement(e);
	} else {
		Atom a = Atom();
		a.setElement(e);
		atoms.push_back(a);
	}
	numEl++;
}

/*
 * Number of atoms currently in the crystal structure
 */
unsigned int CifCrystal::numAtoms() {
	return atoms.size();
}

/*
 * Returns a vector with all atoms
 */
vector<Atom>* CifCrystal::getAtomList() {
	return &atoms;
}

/*
 * Returns a vector with atomic bonds.
 */
vector<CifCrystal::Bond>* CifCrystal::getBondList() {
	return &bonds;
}

/*
 * A check so that all atoms have all properties assigned.
 */
bool CifCrystal::okData() {
	return (numEl == numx) && (numEl == numy) && (numEl == numz)
			&& (numEl == atoms.size());
}

/*
 * Apply each symmetry operation on each atom. If the newly created atom
 * does not already exist (compare coordinates and account for symmetry,
 * this is what Atom::equals does), add it to current Atom list.
 * Afterwards, clear symmetry list.
 * For information about what a symmetry is and how its calculated, see
 * class SymmetryCalculator.
 */
void CifCrystal::processSymmetries() {

	SymmetryCalculator sc = SymmetryCalculator();
	Atom newAtom;
	vector<Atom> newAtoms(0);

	for(unsigned int i = 0; i < symmetries.size(); i++) {//for each symmetry
		//string& sym = symmetries[i];
		if(!sc.setSymmetry(symmetries[i])) { //load the symmetry
			cout << "WARNING: Symmetry number " << i + 1 << " could not be parsed. "
					"Ignoring symmetry." << endl;
			continue;
		}


		for(unsigned int j = 0; j < atoms.size(); j++) {	//for each atom
			sc.loadAtom(atoms[j].getx(),atoms[j].gety(),atoms[j].getz());	//load atom values
			if(!sc.calculate()) { //and calculate
				//something went wrong with the calculations
				cout << "WARNING: Symmetry number " << i + 1 << " could not be calculated."
						" Ignoring symmetry." << endl;
				break; //skip the rest of atoms, because an atom can't be "wrong", only the symmetry
			}
			//now create atom
			newAtom = Atom(atoms[j].getElement(),sc.getx(),sc.gety(),sc.getz());
			newAtom.toCubic();
			bool isNew = true;
			//now cheack if atom already exists
			for(unsigned int k = 0; k < atoms.size();k++) { //check current atoms
				if(newAtom.equals(atoms[k]) ) {
					isNew = false;
					break;
				}
			}
			if(isNew)
				for(unsigned int k = 0; k < newAtoms.size();k++) { //check newly added atoms
					if(newAtom.equals(newAtoms[k]) ) {
					isNew = false;
					break;
				}
			}

			if(isNew) {
				newAtoms.push_back(newAtom);
			}
		}
	}

	//for all new atoms, add to the atom list.
	for(unsigned int k = 0; k < newAtoms.size(); k++) {
		atoms.push_back(newAtoms[k]);
	}

	//clear all the processed symmetries.
	symmetries.clear();

}

/*
 * Creates a supercell. This means, duplicate atom coordinates
 * in x,y and z directions respectivly, so that the new cell
 * (called supercell), has sides a,b,c that are at minimum
 * r_min.
 * E.g. r_min = 20, a = 9. Duplicate into x-direction 3 times,
 * so that new a becomes 27 > 20.
 */
void CifCrystal::makeSuperCell(double r_min) {

	int i = 1; //number of copies.
	vector<Atom> newAtoms; //a holder vector for new atoms.
	while((i*a) < r_min) { //duplicate in x-direction
		for(unsigned int k = 0;k < atoms.size();k++) {
			Atom& at = atoms[k]; //for easier-to-read code
			//now add a new atom, with the same properties as the current one ("at"),
			//but with its x-coordinate incremented with i.
			newAtoms.push_back(Atom(at.getElement(),at.getx()+i,at.gety(),at.getz()));
		}
		i++;
	}
	/*for(unsigned int j = 0;j < newAtoms.size();j++) {
			atoms.push_back(newAtoms[j]);
	}*/ //this is the alternative way to do this
	for(vector<Atom>::iterator it = newAtoms.begin(); it != newAtoms.end();it++) {
		atoms.push_back(*it); //add all atoms from newAtoms to atoms.
	}
	aNew = a*i; //the new a.
	aDup = i;
	newAtoms.clear(); //clear
	i=1; //reset i for next iteration
	while((i*b) < r_min) { //duplicate in y-direction
		for(unsigned int k = 0;k < atoms.size();k++) {
			Atom& at = atoms[k]; //for easier-to-read code
			newAtoms.push_back(Atom(at.getElement(),at.getx(),at.gety()+i,at.getz()));
		}
		i++;
	}
	for(vector<Atom>::iterator it = newAtoms.begin(); it != newAtoms.end();it++) {
		atoms.push_back(*it);
	}
	bNew = b*i; //the new b.
	bDup = i;
	newAtoms.clear();
	i=1;
	while((i*c) < r_min) { //duplicate in z-direction
		for(unsigned int k = 0;k < atoms.size();k++) {
			Atom& at = atoms[k]; //for easier-to-read code
			newAtoms.push_back(Atom(at.getElement(),at.getx(),at.gety(),at.getz()+i));
		}
		i++;
	}
	for(vector<Atom>::iterator it = newAtoms.begin(); it != newAtoms.end();it++) {
		atoms.push_back(*it);
	}
	cNew = c*i; //the new c.
	cDup = i;
}

void CifCrystal::makeTransformations() {
	/*
	 * First, a quick lesson in linear algebra and how this is done, and how
	 * stuff was calculated, because having it only written down on a paper
	 * that I later may or may not find can be bad, in case something will
	 * need to be altered and I won't understand anything here.
	 */

	/*
	 * The atoms in the file are given in relative coordinates, in the
	 * coordinate system of a,b,c, which are vectors in the euclidian system
	 * with lengths a,b,c, and with angles alpha (between b,c), beta (between a,c)
	 * and gamma (between a,b) between them. To not think of the relative thingy
	 * on the coordinates, I see the current coordinates as being written in
	 * the basis of {x',y',z'}, where |x'| == a, |y'| == b, |z'| == c, and with
	 * the given angles. (basically the basis are the vectors a,b,c but technically
	 * a,b,c represents scalars (lengths) and not vectors, so I can't really write
	 * that).
	 *
	 * If I set the x-axis to be oriented along the vector of a, then x' will be
	 * (a,0,0). That is one of the basis vectors for the current coordinate system
	 * (or vector room). Now to y'. See triangle below, which is drawn in x-y plane:
	 *   /|
	 *  / |
	 * /__|
	 *
	 * Bot-left corner is angle gamma, and the hypotenuse length is b. Lets call the
	 * bot side f and right side h. Then we will have b*cos(gamma) = f, b*sin(gamma) = h.
	 * This will give y' = (b*cos(gamma),b*sin(gamma),0).
	 *
	 * With similar geometry, z'.x will be c*cos(beta), and z'.y == c*cos(alpha). It isn't
	 * that easy to calculate z'.z, so I will use the formula (z'.x)²+(z'.y)²+(z'.z)² == c².
	 * Solving for z'.z, I get, in the end, that
	 * z' = (c*cos(beta),c*cos(alpha),c*sqrt(1-cos²(beta)-cos²(alpha))).
	 *
	 * So, in the .cif file the coordinates are given in the basis of
	 * (a,0,0) , (b*cos(gamma), b*sin(gamma),0), (c*cos(beta),c*cos(alpha),c*sqrt(1-cos²(beta)-cos²(alpha))).
	 *
	 * I want to transform it to euclidian basis (1,0,0),(0,1,0),(0,0,1). To do that, I use
	 * the formula B = T*B'. B' is the vector given in the "old" basis, and B are the coordinates
	 * in the new euclidian basis. T is the transformation matrix where its columns are given by
	 * x',y' and z'. (not gonna write it out, too large).
	 *
	 * Esentially, using simple matrix multiplication, this formulas are given, where Ax,Ay,Az are the atom
	 * coordinates as written in the file:
	 *
	 * Nx = Ax*a + Ay*b*cos(gamma) + Ax*c*cos(beta)
	 * Ny = Ay*b*sin(gamma) + Az*c*cos(alpha)
	 * Nz = Az*c*sqrt(1-cos²(beta)-cos²(alpha))
	 *
	 *IMPORTANT IMPORTANT IMPORTANT
	 * NOTE: Transformation matrix is slightly different, the z'.y and z'.z are not the same.
	 * See http://en.wikipedia.org/wiki/Fractional_coordinates for more info. I got a bit
	 * overzealous with these. Using the wiki Transformation matrix.
	 *
	 * NOTE: AFTER more testing, it appears my own transformation matrix gives exact same
	 * result as wikipedia's one... Using my own one just because. And they are simpler.
	 *
	 * Apply these on every atom.
	 */

	double alphar, betar, gammar; //angles in radian.

	//cmath's trigonometric functions need radian angles.
	alphar = alpha*_PI__/180;
	betar = beta*_PI__/180;
	gammar = gamma*_PI__/180;

	//these are some stuff that needs to be calculated many times, might as well
	//do it once here and keep the result.
	const double cosg = cos(gammar);
	const double cosb = cos(betar);
	const double sing = sin(gammar);
	const double cosa = cos(alphar);

	const double zeta = sqrt(1-pow(cosb,2)-pow(cosa,2));
	//const double u = ((cos(alphar)-(cosb*cosg))/(sing));
	//const double v = sqrt(1-pow(cosb,2)-pow(cos(alphar),2)-pow(cosg,2) + (2*cos(alphar)*cosb*cosg));

	//now make the cell cubic, instead of a parallelepiped, so new values
	//of aNew, bNew and cNew are needed, for the makeSmall() function.
	//aNew = aNew; stays the same
	//bNew = bNew * sing; //the y part of the b vector.
	//cNew = cNew * zeta; //the z-part of the c-vector
	//cNew = cNew * v/sing; //the wiki way of doing it.


	for(unsigned int i = 0; i < atoms.size(); i++) {
	 	const double& Ax = atoms[i].getx();
		const double& Ay = atoms[i].gety();
		const double& Az = atoms[i].getz();
		double newx, newy, newz;
		newx = (Ax*a + Ay*b*cosg + Az*c*cosb);
		newy = (Ay*b*sing + Az*c*cosa);
		newz = (Az*c*zeta);
		//newy = (Ay*b*sing + Az*c*u);
		//newz = (Az*c*v/sing);
		atoms[i].setx(newx);
		atoms[i].sety(newy);
		atoms[i].setz(newz);

	}
}

/*
 * Create atomic bonds. An atomic bond should be created between two atoms
 * that have distance less than 2 Å between each other. However, since the
 * cell is repeated many times, atoms may be bound to other atoms outside
 * of the cell. To account for that, bind to the corresponding atom in
 * the current cell.
 */
void CifCrystal::makeBonds() {
	//first initialize some variables

	double alphar, betar, gammar;

	//cmath's trigonometric functions need radian angles.
	alphar = alpha*_PI__/180;
	betar = beta*_PI__/180;
	gammar = gamma*_PI__/180;

	double bx = bNew*cos(gammar);
	double by = bNew*sin(gammar);
	double cx = cNew*cos(betar);
	double cy = cNew*cos(alphar);
	double cz = cNew*sqrt(1-pow(cos(betar),2)-pow(cos(alphar),2));



	for(unsigned int mainInt = 0; mainInt < atoms.size(); mainInt++) { //for each atom
		/*
		 * For each atom, compute the distance to every other atom. If it is
		 * less than 2, create a bond.
		 */
		Atom& a1 = atoms[mainInt];
		for(unsigned int otherInt = 0; otherInt < atoms.size(); otherInt++) {
			if(mainInt!=otherInt) { //no point comparing atom to itself.
				Atom& a2 = atoms[otherInt];
				//now get distance between a1 and a2
				double minDistance = 100;
				for(int i = -1; i < 2; i++) {//from -1 to 1
				for(int j = -1; j < 2; j++) {
				for(int k = -1; k < 2; k++) {
					double tmp = sqrt(
							pow(a1.getx()-a2.getx()+i*aNew+j*bx+k*cx,2) +
							pow(a1.gety()-a2.gety()+j*by+k*cy,2) +
							pow(a1.getz()-a2.getz()+k*cz,2));
					if(tmp < minDistance)
						minDistance = tmp;
				}
				}
				}
				if(minDistance <= 2) {
					//add bond to atom
					a1.addBond(&a2);
					if(mainInt < otherInt) {//dont add duplicate bonds.
						Bond b = {mainInt,otherInt};
						bonds.push_back(b);
					}
				}
			}
		}
	}
}

//for testing purposes
void CifCrystal::printThis() {
	cout.setf(ios_base::fixed);
	cout.precision(4);
	cout << "----- CELL VALUES -----" << endl
		<< "a: ";
	if(isSet[0])
		cout << a << endl;
	else
		cout << "NOT SET" << endl;
	cout << "b: ";
	if(isSet[1])
		cout << b << endl;
	else
		cout << "NOT SET" << endl;
	cout << "c: ";
	if(isSet[2])
		cout << c << endl;
	else
		cout << "NOT SET" << endl;
	cout << "alpha: ";
	if(isSet[3])
		cout << alpha << endl;
	else
		cout << "NOT SET" << endl;
	cout << "beta: ";
	if(isSet[4])
		cout << beta << endl;
	else
		cout << "NOT SET" << endl;
	cout << "gamma: ";
	if(isSet[5])
		cout << gamma << endl;
	else
		cout << "NOT SET" << endl;
	cout << endl;
	cout << "----- SYMMETRIES -----" << endl;
	for(unsigned int i=0;i<symmetries.size();i++) {
		cout << symmetries[i] << endl;
	}
	cout << endl;
	int numO = 0;
	int numSi = 0;
	cout << "----- ATOMS -----" << endl;
	if(!okData()) {
		cout << "BLÄH: " << numEl << ' '
		<< numx << ' ' << numy << ' '
		<< numz << endl;
	}

	for(unsigned int i=0;i<atoms.size();i++) {
		cout << atoms[i].getElement() << i
			<< "\t\t" << atoms[i].getx()
			<< "\t\t" << atoms[i].gety()
			<< "\t\t" << atoms[i].getz() << endl;

		if(atoms[i].getElement() == "O")
			numO++;
		else if(atoms[i].getElement() == "Si")
			numSi++;

	}
	cout << "NumO: " << numO << "\tnumSi: " << numSi;

	cout << endl << endl;
	unsigned int i;
	for(i = 0; i < bonds.size();i++) {
		cout << bonds[i].a1 << ' ' << bonds[i].a2 << endl;
	}
	cout << "NumBounds: " << i << endl;

}


