/*
 * OutputWriter.cpp
 *
 * Copyright Notice: see copyright.txt
 *
 *  Date: 3/8/2012 (D/M/Y)
 *  Author: Dmitrij Lioubartsev
 *  Email: dmitrijl42@gmail.com
 *
 *  For class info, see OutputWriter.h
 */

#include "OutputWriter.h"

#include <fstream>

#include <boost/lexical_cast.hpp>

OutputWriter::OutputWriter(string outputFileName) {
	output = outputFileName;
}

OutputWriter::~OutputWriter() {
	// do nothing
}

void OutputWriter::createXmol(CifCrystal& crystal) {
	removeFileExtension(output);
	output.append(".xmol");
	ofstream xmolStream;
	xmolStream.open(output.c_str());
	vector<Atom>* atomVector = crystal.getAtomList();
	xmolStream << (*atomVector).size() << endl;
	xmolStream << "BOX: " << crystal.getNewa() << '\t'
			<< crystal.getNewb() << '\t' << crystal.getNewc()
			<< '\t' << crystal.get_alpha()
			<< '\t' << crystal.get_beta()
			<< '\t' << crystal.get_gamma() << endl;
	xmolStream.setf(ios_base::fixed);
	xmolStream.precision(4);
	for(unsigned int i = 0; i < crystal.numAtoms(); i++) { //for each atom
		Atom * at = &((*atomVector)[i]);
		xmolStream << (*at).getElement() << ' '
				<< (*at).getx() << ' '
				<< (*at).gety() << ' '
				<< (*at).getz() << endl;
	}
}
/*
void OutputWriter::createSmolDebug(CifCrystal& crystal) {
	removeFileExtension(output);
	output.append(".smol");
	ofstream smolStream;
	smolStream.open(output.c_str());
	vector<Atom>* atomVector = crystal.getAtomList();
	smolStream << (*atomVector).size() << endl;
	smolStream << "# (Atom type) (x) (y) (z) (Q) (typ.)" << endl;
	for(unsigned int i = 0; i < crystal.numAtoms(); i++) { //for each atom
		Atom * at = &((*atomVector)[i]);
		smolStream << (*at).getElement() << ' '
				<< (*at).getx() << ' '
				<< (*at).gety() << ' '
				<< (*at).getz() << ' ';
		if((*at).getElement() == "O") {
			smolStream << "-0.25 O_z" << endl;
		} else {
			smolStream << "+0.5 " << (*at).getElement() <<"_z" << endl;
		}
	}
	vector<CifCrystal::Bond>* bondVector = crystal.getBondList();
	smolStream << (*bondVector).size() << endl;
	for(unsigned int i = 0; i < (*bondVector).size(); i++) { //for each atom
		CifCrystal::Bond& b = ((*bondVector)[i]);
		smolStream << (b.a1 + 1) << ' ' << (b.a2 + 1) << endl;
	}
}
*/
void OutputWriter::createSmol(CifCrystal& crystal) {
	removeFileExtension(output);
	output.append(".smol");
	ofstream smolStream;
	smolStream.open(output.c_str());
	vector<Atom>& atomVector = *(crystal.getAtomList());
	smolStream << atomVector.size() << endl;
	smolStream << "#(Element) (x)        (y)           (z)   "
			"         (Q)       (FF type.)" << endl;
	smolStream.setf(ios_base::fixed);
	smolStream.precision(4);

	for(unsigned int i = 0; i < crystal.numAtoms(); i++) { //for each atom
		Atom& a = atomVector[i];
		writeAndFill(a.getElement(),4,smolStream);
		double tmpval;
		//now print x coordinate
		const int corSpace = 15;
		tmpval = a.getx();
		string s = boost::lexical_cast<string>(tmpval);
		if(abs(tmpval) < 10e-4)
			s = "0";
		else {
			s = boost::lexical_cast<string>(tmpval);
			setPrecision(s);
		}
		writeAndFill(s,corSpace-1,smolStream);
		smolStream << ' ';
		//now print y coordinates
		tmpval = a.gety();
		s = boost::lexical_cast<string>(tmpval);
		if(abs(tmpval) < 10e-4)
			s = "0";
		else {
			s = boost::lexical_cast<string>(tmpval);
			setPrecision(s);
		}
		writeAndFill(s,corSpace-1,smolStream);
		smolStream << ' ';
		//now print z coordinates
		tmpval = a.getz();
		s = boost::lexical_cast<string>(tmpval);
		if(abs(tmpval) < 10e-4) //to fix a stupid boost thingy.
			s = "0";
		else {
			s = boost::lexical_cast<string>(tmpval);
			setPrecision(s);
		}
		writeAndFill(s,corSpace-1,smolStream);
		smolStream << ' ';
		//now print Charge
		tmpval = a.getCharge();
		if(abs(tmpval) < 10e-4) //to fix a stupid boost thingy.
			s = "0";
		else {
			s = boost::lexical_cast<string>(tmpval);
			setPrecision(s);
		}
		writeAndFill(s,9,smolStream);
		smolStream << ' ';
		//now print type
		smolStream << a.getFF_type() << endl;
	}

	smolStream << "###########################################################" << endl;
	smolStream << "# Atomic bonds" << endl;
	smolStream << "###########################################################" << endl;
	vector<CifCrystal::Bond>* bondVector = crystal.getBondList();
	smolStream << (*bondVector).size() << endl;
	for(unsigned int i = 0; i < (*bondVector).size(); i++) { //for each atom
		CifCrystal::Bond& b = ((*bondVector)[i]);
		string tmp = boost::lexical_cast<string>(b.a1);
		writeAndFill(tmp,9,smolStream);
		smolStream << ' ';
		smolStream << (b.a2 + 1) << endl;
	}
}

void OutputWriter::createCar(CifCrystal& crystal, string title) {
	removeFileExtension(output);
	output.append(".car");
	ofstream carStream(output.c_str());
	vector<Atom>& atoms = *(crystal.getAtomList());

	/*
	 * For car file difinition, see
	 * http://hincklab.uthscsa.edu/html/soft_packs/msi_docs/insight980/formats980/File_Formats_1998.html#781840
	 */

	carStream << "!BIOSYM archive 3" << endl; //line 1
	carStream << "PBC=ON" << endl; //line 2
	//note: PBC = periodic something, and this crystal structure is periodic, so it should be on.
	carStream << title << endl;
	carStream << "!DATE" << endl; //leaving date stamp blank
	//Fifth line
	carStream << "PBC";
	typedef  double (CifCrystal::*getMembers)();

	getMembers getters[] = {
			&CifCrystal::getNewa,
			&CifCrystal::getNewb,
			&CifCrystal::getNewc,
			&CifCrystal::get_alpha,
			&CifCrystal::get_beta,
			&CifCrystal::get_gamma
	};
	for(int i = 0; i < 6 ; i++) {
		carStream << ' ';
		double d =  ((crystal).*(getters[i])) ();
		string cor;
		if(abs(d) < 10e-4) //to fix a stupid boost thingy.
			cor = "0";
		else {
			cor = boost::lexical_cast<string>(d);
			setPrecision(cor);
		}
		writeAndFill(cor,9,carStream);
	}
	carStream << " N/A" << endl;

	//now write atom data
	for(unsigned int i = 0; i < atoms.size(); i++) {
		Atom& a = atoms[i];
		string writeThis;
		writeThis = a.getElement();
		writeAndFill(writeThis,5,carStream);
		carStream << ' '; //char nr 6 is empty
		//now write atom coordinates
		//each coordinate has 15 chars
		const int numChars = 14;
		double tmpval;
		//write x
		tmpval = a.getx();
		if(abs(tmpval) < 10e-4) //to fix a stupid boost thingy.
			writeThis = "0";
		else {
			writeThis = boost::lexical_cast<string>(tmpval);
			setPrecision(writeThis);
		}
		writeAndFill(writeThis,numChars,carStream);
		carStream << ' '; //empty char
		//write y
		tmpval = a.gety();
		if(abs(tmpval) < 10e-4) //to fix a stupid boost thingy.
			writeThis = "0";
		else {
			writeThis = boost::lexical_cast<string>(tmpval);
			setPrecision(writeThis);
		}
		writeAndFill(writeThis,numChars,carStream);
		carStream << ' ';
		//write z
		tmpval = a.getz();
		if(abs(tmpval) < 10e-4) //to fix a stupid boost thingy.
			writeThis = "0";
		else {
			writeThis = boost::lexical_cast<string>(tmpval);
			setPrecision(writeThis);
		}
		writeAndFill(writeThis,numChars,carStream);
		carStream << ' ';
		//52-55 Name of residue containing atom.
		carStream << "XXXX"; //4 x:es
		//57-63 Residue sequence number relative to the beginning of the current molecule.
		carStream << " 1      "; //56 empty
		//64-70 Potential function atom type (left justified)
		writeThis = a.getFF_type();
		writeAndFill(writeThis, 7, carStream);
		// 72-73 element symbol
		carStream << ' '; //char 71
		writeThis = a.getElement();
		writeAndFill(writeThis,2,carStream);
		// 75-80 partial charge on atom
		carStream << ' '; //char 74
		tmpval = a.getCharge();
		if(abs(tmpval) < 10e-4) //to fix a stupid boost thingy.
			writeThis = "0";
		else {
			writeThis = boost::lexical_cast<string>(tmpval);
			setPrecision(writeThis);
		}
		writeAndFill(writeThis, 6, carStream);
		carStream << endl;
		//some old stuff:

		//76-80 Absolute atomic sequence number relative to the
		//beginning of the entire molecular system.
		//carStream << " 0.000" << endl;
	}
	carStream << "end" << endl << "end";
}

void OutputWriter::printCrystal(CifCrystal& crystal) {
	crystal.printThis();
}

void OutputWriter::removeFileExtension(string& fileName) {
	unsigned int i = fileName.length()-1;
	while(fileName[i] != '.' && i > 0) {
		i--;
	}

	if(i != 0) {
		fileName.erase(i,fileName.length()-i);
	}
}

/*
 * Similar to removeFileExtension, but this keeps 'precision'
 * amount of characters after the decimal point. Used for
 * printing number string with a certain length.
 * (since ofstream::precision only works for numeric)
 */
void OutputWriter::setPrecision(string& doubl, int precision) {
	//first find the .
	int i = doubl.length()-1;
	while(doubl[i] != '.' && i >= 0) {
		i--;
	}

	if(i < 0) //no decimal point, do nothing
		return;

	// i is now the position of the decimal point

	if(((int) (doubl.length())) - i - 1 > precision) {
		int breaker = i+precision+1; //+1 accounts for the .
		doubl.erase(breaker, doubl.length()-breaker);
	} //else do nothing

	return;
}

/*
 * Writes the string writeThis to the stream. The amount of characters written
 * equals to 'space'. If writeThis.length is shorter, fill rest with space characters.
 */
void OutputWriter::writeAndFill(string writeThis, int space, ofstream& stream) {

	if(((int)(writeThis.length())) > space) //casted to supress warnings
		stream << writeThis.substr(0,space);
	else {
		stream << writeThis;
		int numSpaces = space-writeThis.length();
		for(int i = 0; i < numSpaces; i++) stream << ' ';
	}
}
