/*
 * SubAtomHolder.h
 *
 * Copyright Notice: see copyright.txt
 *
 * Date: 3/8/2012 (D/M/Y)
 * Author: Dmitrij Lioubartsev
 * Email: dmitrijl42@gmail.com
 *
 * This is the main class for substituting atoms according to input data.
 * The algorithm has 6 main steps, after initializing basic stuff.
 *
 * 1. From given ratios, create atomsLeft, which holds number of atoms to be
 * 			substituted of each element.
 * 2. Set distance of all atoms, by number of bonds, to a source atom.
 * 3. Sort atom vector by this distance.
 * 4. Give all elements a priority number, based on number of occurences in exlude strings.
 * 5. Go through each atomn and substitue with an atom that meets these criteria:
 * 		a) Has highest priority
 * 		b) Does not violate excludes
 * 		c) Does not have atomsLeft == 0 (after a substitution, atomsLeft for this element
 * 			 is decreased by 1)
 * 6. Shuffle up, to create randomness.
 */

#ifndef SUBATOMHOLDER_H_
#define SUBATOMHOLDER_H_

#include <iostream>
#include <vector>
#include <map>
#include <string>
#include <queue>
#include <algorithm>
#include <cmath>
#include <stdexcept>
using namespace std;

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int_distribution.hpp>

#include "Atom.h"

class Substitution {
private: //fields
	vector<Atom*> atoms; //all relevant atoms to be substituted
	vector<pair<string,double> >* ratios;	//ratios
	vector<Exclusion> excludes;				//excludes
	map<string,int> subsLeft;				//number of atoms left from various elements to substitute
	map<string,int> subsLeftcpy;			//will be initialized to subsLeft but never change values.
											//it is not important for the algorithm, but it is helpful
											//for output writing.
	vector<pair<string,int> > priorities;	//a vector of priorities

	int totalAtomCount; // == atoms.size()

	/*
	 * A compare class used to sort atoms by their value distToSrc
	 */
	struct AtomCmpDist {
		bool operator()(const Atom * a1,const Atom * a2 ) const {
			return (*a1).getDistToSrc() < (*a2).getDistToSrc();
		}
	} compDist;

	/*
	 * A compare class used to sort priorities by highest priority first.
	 */
	struct compByPrio {
		bool operator()(const pair<string,int>& p1, const pair<string,int>& p2) const {
			return p1.second > p2.second; //largest first
		}
	} prioComp;

private: //methods
	bool initSubsLeft();  //step 1
	//calculates subsLeft, only called once from constructor.
	/*
	 * Sets the distance off all atoms to the source (atoms[0]), and distance as in
	 * distance by bonds. Using a BFS algorithm.
	 */
	void setDistToSrcAll(); //step 2

	/*
	 * Initializes the priorities vector.
	 */
	void initPriorities(); //step 4
	bool subAll(); //step 5

	/*
	 * Step 6, swaps elements on two atoms if its legalSubstitute returns yes.
	 * Loops for seed times.
	 */
	void shuffle(int seed);

	bool legalSubstitute(Atom& a, string sub, bool checkZero = true);

public:
	Substitution(vector<Atom>* substitutes, vector<pair<string,double> >* ratios, vector<string>* excludes);
	virtual ~Substitution();

	bool substitute();

	void printAtomsSubstituted();


};

#endif /* SUBATOMHOLDER_H_ */
