/*
 * SymmetryCalculator.cpp
 *
  * Copyright Notice: see copyright.txt
 *
 * Date: 3/8/2012 (D/M/Y)
 * Author: Dmitrij Lioubartsev
 * Email: dmitrijl42@gmail.com
 *
 *For class info, see SymmetryCalculator.h
 */

#include "SymmetryCalculator.h"
#include <string>
#include <cctype>
#include <cstdlib>
using namespace std;

SymmetryCalculator::SymmetryCalculator() {
	xyz[0] = xyz[1] = xyz[2] = ""; //each symmetry string
	xyzNVal[0] = xyzNVal[1] = xyzNVal[2] = 0;
}

SymmetryCalculator::~SymmetryCalculator() {
	// do nothing
}

// some macros for easier coding
#define OPERATOR(x)  (x == '+' || x == '-' || x == '/' || x == '*')
#define DIGIT(x) ((x >= 48 && x <= 57) || x == '.')
#define VARIABLE(x) (x == 'x' || x == 'y' || x == 'z')
#define LAST_CHAR ((xyz[dim])[xyz[dim].length()-1])

/*
 * Returns false if symmetry is bad (wrong syntax etc.)
 */
bool SymmetryCalculator::setSymmetry(const string& sym) {

	//repair symmetry, do the following:
	/*
	 * 1. split into dimensions
	 * 2. delete whitespace characters
	 * 3. change empty dimensions into x,y,z respectivly
	 * 4. insert '*' into expressions, eg "2x" -> "2*x".
	 */

	short dim = 0; //x,y,z
	bool ok = true;
	unsigned short numLP = 0; //for a quick check for parenthesis
	unsigned short numRP = 0;
	xyz[0] = xyz[1] = xyz[2] = ""; //reset old symmetries

	for(unsigned int k = 0; k < sym.length(); k++) { //go through all chars in sym
		char c = sym[k];
		if(DIGIT(c)) {								 //if 0-9
			char ch = LAST_CHAR;
			if(ch == ')' || VARIABLE(ch))
				xyz[dim].append(1,'*');
			xyz[dim].append(1,c);
		} else if(VARIABLE(c)) { 					// if x,y,z
			char ch = LAST_CHAR;
			if(ch == ')' || VARIABLE(ch) || DIGIT(ch))
				xyz[dim].append(1,'*');
			xyz[dim].append(1,c);
		} else if(OPERATOR(c)) { 						// if '('
			xyz[dim].append(1,c);
		} else if(c == '(') { 						// if '('
			char ch = LAST_CHAR;
			if(ch == ')' || VARIABLE(ch) || DIGIT(ch))
				xyz[dim].append(1,'*');
			xyz[dim].append(1,c);
			numLP++;
		} else if(c == ')') { 						// if ')'
			if(numLP > numRP) {
				xyz[dim].append(1,c);
				numRP++;
			} else {
				ok = false;
				break;
			}
		} else if(c == ',') {						// if ','
			if(dim > 1) { //NOPENOPENOPENOPE
				ok = false;
				break;
			}
			dim++;
		} else if(c == ' ' || c == '\n' || c == '\t') { //whitespace
			//do nothing
		} else {
			ok = false;							 //unknown/not allowed character
			break;
		}
	}
	//fill dimensions not written
	if(dim == 0) {
		xyz[1] = "y";
		xyz[2] = "z";
	} else if(dim == 1) {
		xyz[2] = "z";
	}
	//fill empty dimensions
	if(xyz[0] == "") {
		xyz[0] = "x";
	}
	if(xyz[1] == "") {
		xyz[1] = "y";
	}
	if(xyz[2] == "") {
		xyz[2] = "z";
	}
	//check parenthesis
	if(numLP != numRP)
		ok = false;

	return ok;
}

/*
 * Load atom values.
 */
void SymmetryCalculator::loadAtom(double x, double y, double z) {
	xyzOVal[0] = x; xyzOVal[1] = y; xyzOVal[2] = z;
}

/*
 * This below is the old function, which replace all instances of
 * x, y and z with their respective specified atom values.
 * It is not used because there is a better way of doing it, but its
 * good to keep old stuff just in case.
 */

	//parse double values
	/*
	string xstr = boost::lexical_cast<string>(x);
	string ystr = boost::lexical_cast<string>(y);
	string zstr = boost::lexical_cast<string>(z);

	for(int dim = 0; dim < 3;dim++) //for all three indices of xyz.
	for(int i = 0; i < xyz[dim];i++) {
		if((xyz[dim])[i] == 'x') {
			xyz[dim].replace(i,1,xstr);
			i += xstr.length()-1;
		} else if((xyz[dim])[i] == 'y') {
			xyz[dim].replace(i,1,ystr);
			i += ystr.length()-1;
		} else if((xyz[dim])[i] == 'z') {
			xyz[dim].replace(i,1,zstr);
			i += zstr.length()-1;
		}//else do nothing
	}
*/



double SymmetryCalculator::getx() { return xyzNVal[0];}
double SymmetryCalculator::gety() { return xyzNVal[1];}
double SymmetryCalculator::getz() { return xyzNVal[2];}

//calculation methods

//the public interface method
/*
 * Returns false if could not calculate (something wrong with symmetry syntax)
 */
bool SymmetryCalculator::calculate() {

	//calculate each expression
	res xres = calculateTerms(xyz[0],0);
	res yres = calculateTerms(xyz[1],0);
	res zres = calculateTerms(xyz[2],0);

	//set results
	xyzNVal[0] = xres.result;
	xyzNVal[1] = yres.result;
	xyzNVal[2] = zres.result;

	//if bad results, return false
	if(xres.ip == -1 || yres.ip == -1 || zres.ip == -1)
		return false;
	else
		return true;
}

//calculation methods

/*
 * Calculates terms
 */
SymmetryCalculator::res SymmetryCalculator::calculateTerms(string& expr, int expp) {

	res leftSide;

	if(expp == -1 || expp >= (int) expr.length()) {
		leftSide.result = 0;
		leftSide.ip = -1;
		return leftSide;
	}

	leftSide = calculateProducts(expr,expp);
	while(leftSide.ip < (int)expr.length() && (expr[leftSide.ip] == '+' || expr[leftSide.ip] == '-')) {
		//if first expression in while loop header is false, the other two aren't checked,
		//and thus no index out of bound errors should occur.
		res rightSide = calculateProducts(expr,leftSide.ip + 1); //+1 skips the operator
		if(expr[leftSide.ip] == '+') {
			leftSide.result += rightSide.result;
			leftSide.ip = rightSide.ip;
		} else { //subtract
			leftSide.result -= rightSide.result;
			leftSide.ip = rightSide.ip;
		}
	}

	return leftSide;
}

/*
 * calculate products.
 */
SymmetryCalculator::res SymmetryCalculator::calculateProducts(string& expr, int expp) {

	res leftSide;

	if(expp == -1 || expp >= (int) expr.length()) {
		leftSide.result = 0;
		leftSide.ip = -1;
		return leftSide;
	}


	leftSide = getValue(expr,expp);
	while(leftSide.ip < (int)expr.length() && (expr[leftSide.ip] == '*' || expr[leftSide.ip] == '/')) {
		//if first expression in while loop header is false, the other two aren't checked,
		//and thus no index out of bound errors should occur.
		res rightSide = getValue(expr,leftSide.ip + 1); //+1 skips the operator
		if(expr[leftSide.ip] == '*') {
			leftSide.result *= rightSide.result;
			leftSide.ip = rightSide.ip;
		} else { //subtract
			leftSide.result /= rightSide.result;
			leftSide.ip = rightSide.ip;
		}
	}

	return leftSide;
}

/*
 * Fetches the next value.
 */
SymmetryCalculator::res SymmetryCalculator::getValue(string& expr, int expp) {

	res val;

	if(expp == -1 || expp >= (int)expr.length())  {
		val.result = 0;
		val.ip = -1;
		return val;
	} //if called with a bad argument or error argument


	char& cs = expr[expp]; //current symbol, this is for shorter and less messy code

	switch(cs) {
	 //std::atof() can handle expressions like "+2.35" or "-2.35", but this is
		//an easy way to also cover the cases like -x, +x or -(expr).
	case '-' :
		val = getValue(expr,expp+1);
		val.result *= -1;
		break;
	case '+' :
		val = getValue(expr,expp+1);
		break;
	case '(' : //parenthesis. Use recursion on the inner value.
		val = calculateTerms(expr,expp+1);
		val.ip += 1;
		break;
	case '.' ://if its a number. Normal digits and expressions
				//like ".5" are allowed.
	case '0' : case '1' : case '2' : case '3' : case '4' :
	case '5' : case '6' : case '7' : case '8' : case '9' :
	{
		int exppcpy = expp;
		while((!(OPERATOR(expr[expp]))) && expp < (int)expr.length() && expr[expp] != ')')
			expp++;
		val.result = atof(expr.substr(exppcpy,exppcpy-expp).c_str());
		val.ip = expp;
		break;
	}

		//here follows the variable expressions. There will not be cases like "3x",
		//because when the symmetry string is first loaded, it will be "repaired"
		//to "3*x", so there should never be an expression which has a variable
		//letter and is longer than one char.
	case 'x' :
		val.result = xyzOVal[0];
		val.ip = expp + 1;
		break;
	case 'y' :
		val.result = xyzOVal[1];
		val.ip = expp + 1;
		break;
	case 'z' :
		val.result = xyzOVal[2];
		val.ip = expp + 1;
		break;
	default: //something unrecognizable, return the error symbol (-1)
		val.result = 0;
		val.ip = -1;
		break;
	}

	return val;
}

#undef OPERATOR
#undef DIGIT
#undef VARIABLE
#undef LAST_CHAR

